                       Appendix  B

             Specification of data of real time clock
             -----------------------------------------
1. Introduction:
   
   The AT BIOS uses nonvolatile CMOS RAM to store Real Time Clock (RTC),
   System configuration, System diagnostic, and other informaton.

2. Summary:

Ŀ
      Index       Definition                                             
-------------------------------------------------------------------------
        0h        Seconds                                                
        1h        Second Alarm                                           
        2h        Minutes                                                
        3h        Minute Alarm                                           
        4h        Hours                                                  
        5h        Hour Alarm                                             
        6h        Day of Week                                            
        7h        Date of Month                                          
        8h        Month                                                  
        9h        Year                                                   
       0Ah        Status Register A                                      
       0Bh        Status Register B                                      
       0Ch        Status Register C                                      
       0Dh        Status Register D                                      
       0Eh        Diagnostic Status Byte                                 
       0Fh        Shut Down Status Byte                                  
       10h        Diskette Drive Type Byte                               
       11h        Reserved                                               
       12h        Fixed Disk Type Byte                                   
       13h        Reserved                                               
       14h        Equipment Byte                                         
       15h        Low Base Memory Bytes                                  
       16h        High Base Memory Bytes                                 
       17h        Low Memory Expansion Bytes                             
       18h        High Memory Expansion Bytes                            
       19h        Drive C Extended Byte                                  
       1Ah        Drive D extended byte                                  
       1Bh        Hard disk power inactive time (for laptop)             
       1Ch        LCD power inactive time (for laptop)                   
       1Dh        System power inactive time (for laptop)                
       1Eh-2Dh    Reserved for User-defined Fix Disk Type                
       2Eh        High Byte of Checksum of CMOS                          
       2Fh        Low Byte of Checksum of CMOS                           
       30h        Low Byte of expansion memory size                      
       31h        High Byte of expansion memory size                     

       32h        Date Century Byte                                      
       33h        Information Flag                                       
       34h        COM1 status                                            
       35h        MODEM status                                           
       36h        Reserved                                               
       37h        Parallel/Serial/EMS port status                        
       38h        Reserved                                               
       39h        Miscellaneous 1                                        
       3Ah,3Bh    Reserved                                               
       3Ch-3Fh    Password area                                          


Ŀ
00h -- Seconds                                                           
01h -- Second Alarm                                                      
02h -- Minutes                                                           
03h -- Minute Alarm                                                      
04h -- Hours                                                             
05h -- Hour Alarm                                                        
06h -- Day of Week                                                       
07h -- Date of Month                                                     
08h -- Month                                                             
09h -- Year                                                              
0Ah -- Status Register A                                                 
        bit 7   -- Update in Progress (UIP)                              
                    0 = Current date/time available                      
                    1 = Time update in progress                          
        bit 6-4 -- 22-Stage Divider (DV2-DV0)                            
                    010 = System initializes to select a 32.768kHz time  
                          base frequency                                 
        bit 3-0 -- Rate Selection Bits (RS3-RS0)                         
                    110 = System initializes to select a 1.024 kHz square
                          wave output frequency and a 976.562 microsecond
                          periodic interrupt rate                        
                                                                         
0Bh -- Status Register B                                                 
        bit 7   -- Set                                                   
                    1 = initializes the 14 time-bytes and                
                        aborts any update cycle.                         
                    0 = updates the cycle normally by advancing          
                        the counts at one-per-second.                    
        bit 6   -- Periodic Interrupt Enable (PIE) (Read/Write)          
                    1 = enable interrupt at a rate specified             
                        by the divider bit in Register A.                
                    0 = disable interrupt         (DEFAULT)              
        bit 5   -- Alarm Interrupt Enable (AIE)                          
                    1 = enables the alarm interrupt                      
                    0 = disables                 (DEFAULT)               
        bit 4   -- Update-Ended Interrupt Enabled (UIE)                  
                    1 = enables the UIE                                  
                    0 = disables                 (DEFAULT)               
        bit 3   -- Square Wave Enabled (SQWE)                            
                    1 = enables the square-wave frequency as set by      
                        register A                                       
                    0 = disables                 (DEFAULT)               
        bit 2   -- Date/Time Mode (DM)                                   
                    1 = Binary format                                    
                    0 = BCD format                (DEFAULT)              
        bit 1   -- 24/12-hour mode                                       
                    1 = 24-hour mode              (DEFAULT)              
                    0 = 12-hour mode                                     

        bit 0   -- Daylight Savings Enabled (DSE)                        
                    1 = enable daylight savings                          
                    0 = disable it                (DEFAULT)              
                                                                         
0Ch -- Status Register C                                                 
        bit 7-4 -- IRQF, PF, AF, UF flag bits      (READ ONLY)           
                   are affected respectively when the PIE, AIE, and      
                   UIE interrupts are enabled in register B.             
        bit 3-0 -- Reserved = 0                                          
                                                                         
0Dh -- Status Register D                                                 
        bit 7   -- Valid RAM Bit (VRB)             (READ ONLY)           
                    1 = RTC power good                                   
                    0 = RTC has lost power                               
        bit 6-0 -- Reserved = 0                                          
                                                                         
0Eh -- Diagnostic Status Byte                                            
        bit 0,1 -- Reserved = 0                                          
        bit 2   -- Time Status Indicator                                 
                    1 = invalid                                          
                    0 = valid                                            
        bit 3   -- Fixed disk controller/Drive C initialization          
                    1 = FAILED, prevents system from booting up          
                    0 = Functions properly - can boot up                 
        bit 4   -- Memory size comparison                                
                    1 = memory size record different from power-on check 
                    0 = memory size is the same                          
        bit 5   -- Incorrect configuration                               
                    1 = invalid                                          
                    0 = valid                                            
        bit 6   -- Configuration status                                  
                    1 = Checksum incorrect                               
                    0 = Checksum ok                                      
        bit 7   -- power status                                          
                    1 = battery dead                                     
                    0 = battery ok                                       
                                                                         
0Fh -- Shutdown Status Byte   (for POST usage)                           
                                                                         

10h -- Diskette Drive Type Byte                                          
        bit 7-4 -- Type of 1st diskette drive                            
                    0000 = No drive present                              
                    0001 = Double-sided, 48 TPI, 360K                    
                    0010 = High capacity, 96 TPI, 1.2M                   
                    0011 = 720K disk drive                               
                    0100 = 1.44M disk drive                              
                    0111~1111 = Reserved                                 
        bit 3-0 -- Type of 2nd diskette drive                            
                    0000 = No drive present                              
                    0001 = Double-sided, 48 TPI, 360K                    
                    0010 = High capacity, 96 TPI, 1.2M                   
                    0011 = 720K disk drive                               
                    0100 =  1.44M disk drive                             
                    0111~1111 = Reserved                                 
                                                                         
11h -- Reserved = 0                                                      
                                                                         
12h -- Fixed Disk Type Byte                                              
        bit 7-4 -- Define Fixed Disk Type of Drive C                     
                    0000 = No fixed disk drive is present                
                    0001~1110 = Define type 1 thru 14                    
                    1111 = Type 16 thru 255 see "Drive C extended        
                           byte" (19h)                                   
        bit 3-0 -- Define Fixed Disk Type of Drive D                     
                    0000    No fixed disk drive is present               
                    0001~1110 = Define type 1 thru 14                    
                    1111 = Type 16 thru 255,see "Drive D extended        
                           byte" (1Ah)                                   
           Ŀ     
                                    Write   Landing  Sectors      
            Type  Cylinder  Heads  Precomp   Zone     /Track      
           Ĵ     
              1      306      4      128      305       17        
              2      615      4      300      615       17        
              3      615      6      300      615       17        
              4      940      8      512      940       17        
              5      940      6      512      940       17        
              6      615      4    0FFFFH     615       17        
              7      462      8      256      511       17        
              8      733      5    0FFFFH     733       17        
              9      900     15    0FFFFH     901       17        
             10      820      3    0FFFFH     820       17        
             11      855      5    0FFFFH     855       17        
             12      855      7    0FFFFH     855       17        
             13      306      8      128      319       17        
             14      733      7    0FFFFH     733       17        
           Ĵ     
             15   Extended type, see byte 19h & 1Ah                   
                

                                                                         
13h -- Reserved = 0                                                      
                                                                         
14h -- Equipment Byte                                                    
        bit 7,6 -- Number of diskette drives installed                   
                    0 0 = One drive                                      
                    0 1 = Two drive                                      
                    1 0 = Reserved                                       
                    1 1 = Reserved                                       
        bit 5,4 -- Primary display                                       
                    0 0 = EGA  (Adapter has its own BIOS)                
                    0 1 = CGA in 40-column mode                          
                    1 0 = CGA in 80-column mode                          
                    1 1 = Monochrome                                     
        bit 3,2 -- Reserved                                              
        bit 1   -- Math corprocessor                                     
                    1 = installed                                        
                    0 = not installed                                    
        bit 0   -- Diskette drive                                        
                    1 = installed                                        
                    0 = Not any installed                                
                                                                         
15h -- Low  Base Memory Bytes                                            
16h -- High Base Memory Bytes                                            
        NOTE:  16H-15H store the amount of memory installed below the 1  
               MB address space (in 1KB unit), 3C00h maximum.            
                                                                         
17h -- Low  Memory Expansion Bytes                                       
18h -- High Memory Expansion Bytes                                       
        NOTE:  18H-17H store the amount of memory installed above the 1  
               MB address space (in 1KB unit).                           
                                                                         
19h -- Drive C Extended Byte                                             
        00000000 thru 00001111 are reserved                              
        00010000 thru 11111111 define type 16 thru 255                   

1Ah -- Drive D Extended Byte                                             
        00000000 thru 00001111 are reserved                              
        00010000 thru 11111111 define type 16 thru 255                   
                                                                         
           Ŀ     
                                    Write   Landing  Sectors      
            Type  Cylinder  Heads  Precomp   Zone     /Track      
           Ĵ     
             16      612      4        0      663       17        
             17      977      5      300      977       17        
             18      977      7    0FFFFH     977       17        
             19     1024      7      512     1023       17        
             20      733      5      300      732       17        
             21      733      7      300      732       17        
             22      733      5      300      733       17        
             23      306      4        0      336       17        
             24      612      4      305      663       17        
             25      306      4    0FFFFH     340       17        
             26      612      4    0FFFFH     670       17        
             27      698      7      300      732       17        
             28      976      5      488      977       17        
             29      306      4        0      340       17        
             30      611      4      306      663       17        
             31      732      7      300      732       17        
             32     1023      5    0FFFFH    1023       17        
           33..59 Reserved                                        
             60     1024      9    0FFFFH    1023       17        
             61     1224      7    0FFFFH    1223       17        
             62     1224     11    0FFFFH    1223       17        
             63     1224     15    0FFFFH    1223       17        
             64     1024      8    0FFFFH    1023       17        
             65     1024     11    0FFFFH    1023       17        
             66      918     11    0FFFFH    1023       17        
             67      925      9    0FFFFH     926       17        
             68     1024     10    0FFFFH    1023       17        
             69     1024     12    0FFFFH    1023       17        
             70     1024     13    0FFFFH    1023       17        
             71     1024     14    0FFFFH    1023       17        
             72     1024      2    0FFFFH    1023       17        
             73     1024     16    0FFFFH    1023       17        
             74      918     15    0FFFFH    1023       17        
             75      820      6    0FFFFH     820       17        
             76     1024      5      512     1023       17        
             77     1024      8      512     1023       17        
             78     1224      7    0FFFFH    1223       35        
             79     1224     11    0FFFFH    1223       35        
             80     1224     15    0FFFFH    1223       35        
             81     1024      7    0FFFFH    1223       17        

             82     1024     11    0FFFFH    1223       17        
             83     1024     15    0FFFFH    1223       17        
             84      776      8    0FFFFH     775       33        
             85      926     13    0FFFFH     926       17        
             86      805      4    0FFFFH     805       26        
             87      976      5    0FFFFH     976       17        
             88      745      4    0FFFFH     745       28        
             89      747      4    0FFFFH     760       28        
           Ĵ     
           90-254 Reserved                                            
           Ĵ     
            255       User-Defined Type, see below                    
                
                                                                         
                                                                         
1Bh -- Hard disk power inactive time (for laptop)                        
                                                                         
1Ch -- LCD power inactive time (for laptop)                              
                                                                         
1Dh -- System power inactive time (for laptop)                           
                                                                         
1Eh-2Dh -- Reserved for User-defined Fixed Disk Type                     
            For drive C :                                                
                          1Eh     Sectors per track                      
                          20h     High byte of cylinder #                
                          21h     Low  byte of cylinder #                
                          22h     Head #                                 
                          23h     High byte of write pre-comp            
                          24h     Low  byte of write pre-comp            
                          25h     High byte of landing-zone              
                          26h     Low  byte of landing-zone              
            For drive D :                                                
                          1Fh     Sectors per track                      
                          27h     High byte of cylinder #                
                          28h     Low  byte of cylinder #                
                          29h     Head #                                 
                          2ah     High byte of write pre-comp            
                          2bh     Low  byte of write pre-comp            
                          2ch     High byte of landing-zone              
                          2dh     Low  byte of landing-zone              
                                                                         
2Eh -- High Byte of Checksum                                             
2Fh -- Low  Byte of Checksum                                             
        NOTE:  Checksum is calculated at address 10h to 2Dh.             

30h -- Low-byte  Expansion Memory Size                                   
31h -- High-byte Expansion Memory Size                                   
        NOTE:  These bytes reflect the total expansion memory            
               size above the 1MB as determined by POST,                 
               3C00h is the maximum and can be accessed by               
               INT 15h function 88h.                                     
                                                                         
32h -- Date Century Byte -- BCD value for the century                    
                                                                         
33h -- Information Flag                                                  
        bit 7   -- When set, this bit indicates that the top             
                   128K of base memory is installed.                     
        bit 6   -- This bit is set to instruct the Setup utility to      
                   display the first user message after the initial setup
        bit 5-0 -- Reserved                                              
                                                                         
34h -- COM1 status                                                       
        bit 7-5 -- baud rate                                             
                    000 =  110                                           
                    001 =  150	                                          
                    010 =  300                                           
                    011 =  600                                           
                    100 = 1200                                           
                    101 = 2400                                           
                    110 = 4800                                           
                    111 = 9600                                           
        bit 4,3 -- parity	                                          
                    0 0 = None 	                                  
                    0 1 = odd 	                                          
                    1 0 = none                                           
                    1 1 = even 	                                  
        bit 2   -- stop bit                                              
                    0 = 1 bit                                            
	            1 = 2 bit                                             
        bit 0,1 -- data length                                           
               1 0 = 7 bits                                              
                    1 1 = 8 bits                                         
                                                                         
35h -- MODEM status                                                      
        bit 7-2 -- Reserved                                              
        bit 1   -- FC0000 - FFFFFF memory status                         
                    0 = disabled                                         
                    1 = enabled                                          
        bit 0   -- MODEM power on/off                                    
                    0 = power on                                         
                    1 = power off                                        
36h -- Reserved                                                          

37h -- Miscellaneous 0                                                   
        bit 7   -- EMS status                                            
                    0 = disabled                                         
                    1 = enabled                                          
        bit 6-4 -- EMS size                                              
       	bit 3   -- Serial port 1                                  
                    0 = disabled                                         
                    1 = enabled                                          
       	bit 2   -- Serial port 2                                  
                    0 = disabled                                         
                    1 = enabled                                          
        bit 1,0 -- Parallel port                                         
                    0 0 = parallel port 1 (3BCh)                         
                    0 1 = parallel port 2 (378h)                         
                    1 0 = parallel port 3 (278h)                         
                    1 1 = disabled                                       
                                                                         
38h -- Reserved                                                          
                                                                         
39h -- Miscellaneous 1                                                   
        bit 7   -- Cache                                                 
                    0 = cache off                                        
                    1 = cache on                                         
        bit 6-4 -- Speed                                                 
                    111= top speed                                       
                    000= smart                                           
                                                                         
        bit 3   -- FDD write protect                                     
                    0 = normal                                           
                    1 = protected                                        
        bit 2   -- WDD write protect                                     
                    0 = normal                                           
                    1 = protected                                        
        bit 1   -- Video RAM BIOS                                        
                    0 = disabled                                         
                    1 = enabled                                          
        bit 0   -- System RAM BIOS                                       
                    0 = disabled                                         
                    1 = enabled                                          
                                                                         
Note:   Bit 4-6 provide 8 speed levels.                                  
        000 represents the smart speed.                                  
        111 represents the highest speed.                                
        001-110 are machine dependent.                                   

                                                                         
3Ah-3Bh -- Reserved                                                      
                                                                         
3Ch-3Fh -- Password                                                      
    3Ch,3Dh -- CRC check word with password                              
    3Eh     -- check sum byte with password                              
    3Fh     -- check sum byte with 3ch - 3dh                             
                                                                         
                                                                         
                                                                         
NOTE : The status bytes in CMOS original now is placed in BIOS data area.
       The new definition of BIOS data area is described below.          
                                                                         
40:C7 -- Write Protect service routine                                   
                                                                         
    bios_ds       dw  --  save user DS                                   
    vacc_off      dw  --                                                 
    vacc_seg	dw  --                                                    
    bios_vac	db  --  for BIOS vaccine                                  
                            bit 0   -- WDD write Protect                 
                                         0 = normal                      
                                         1 = write protected                                           
                            bit 1   -- FDD write Protect                 
                                         0 = normal                      
                                         1 = write protected             
                            bit 2   -- WDD Write Protect message         
                                         0 = normal                      
                                         1 = write protected             
                            bit 3   -- FDD Write Protect message         
                                         0 = normal                      
                                         1 = write protected             
                            bit 4-7 -- Reserved                          
    day_count     dw  --  day count                                      
                                                                         
                                                                         
40:D0 -- System speed status and Cache status                            
                                                                         
    status_speed  db  --  speed status                                   
                            bit 0   -- console status                    
                                         0 = console exists              
                                         1 = console absent              
                            bit 1   -- RAM VIDEO status                  
                                         0 = RAM VIDEO ok                
                                         1 = RAM VIDEO bad               
                            bit 2   -- RAM BIOS status                   
                                         0 = RAM BIOS ok                 
                                         1 = RAM BIOS bad                

                            bit 3   -- cache status (1100/25/20C only)   
                                         0 = cache off                   
                                         1 = cache on                    
                            bit 4   -- top speed                         
                                         for 1100/25/20C                 
                                           0 = 1100/20C                  
                                           1 = 1100/25                   
                                         for MSAT/16/20                  
                                           0 = MSAT/16                   
                                           1 = MSAT/20                   
                            bit 5   -- Weitek 3167 Math coprocessor      
                                           0 = absent                    
                                           1 = present                   
                            bit 6,7 -- speed status                      
                                         1 1 = smart                     
                                         1 0 = top speed                 
                                         0 1 = reserved                  
                                         0 0 = none of above             
    bios_speed    db --                                                  
                            bit 0   -- RAM BIOS status                   
                                         0 = disables RAM BIOS           
                                         1 = enables RAM BIOS            
                            bit 1   -- RAM BIOS write                    
                                         0 = disables RAM BIOS write     
                                         1 = enables RAM BIOS write      
                            bit 2   -- speed select                      
                                         0 = disables speed change       
                                         1 = enables speed change        
                            bit 3   -- cache status                      
                                         0 = disables cache system       
                                         1 = enables cache system        
                            bit 4   -- flush cache                       
                                         0 = normal                      
                                         1 = flush cache memory          
                            bit 5-7  speed parameter (valid when bit 2   
                                           = 1)                          
                                         000 = 6 MHZ                     
                                         100 = 8 MHZ                     
                                         xxx = bigger value, more fast   
    bios_d        db --                                                  
                            bit 0   -- start status                      
                                         0 = jump F000:FFF0H             
                                         1 = cold start                  
                            bit 1   -- reversed                          
                            bit 2   -- cache status                      
                                         0 = cache ok                    
                                         1 = cache bad                   
                            bit 3   -- reversed                          

                            bit 4   -- WDD status                        
                                         0 = 1st WDD active              
                                         1 = 2nd WDD active              
                            bit 5   -- connor WDD                        
                                         0 = non CONNER                  
                                         1 = CONNER WDD for 1st WDD      
                            bit 6   -- connor WDD                        
                                         0 = non CONNER                  
                                         1 = CONNER WDD for 2nd WDD      
                            bit 7   -- setup status                      
                                         0 = not in setup                
                            	          1 = in setup                    
                                                                         

