               the Specification of System Communication
               -----------------------------------------

function:

	ah = 000h Initialize the communications port
	ah = 001h Send character
	ah = 002h Receive a character
	ah = 003h Return the communications port status
	ah = 004h Extended initialize
	ah = 005h Extended communications port control
	ah = 006h - 0ffh - reserved - null return

Note:
 1. Function 4 and 5 are only for PS/2 products.

input:
	ah - function number
	dx - communications port number (0,1,2,3)

Note:
 1. No checking for communications port number.
    If port address is 0, then null return.

output:
	see individual function comment.

reference BIOS data area:
	     serial port address
	    ---------------------
	000h -- word -- COMM1 port address
	002h -- word -- COMM2 port address
	004h -- word -- COMM3 port address
	006h -- word -- COMM4 port address
	     serial port timeout value
	    ---------------------------
	07ch -- byte -- COMM1 timeout value
	07dh -- byte -- COMM2 timeout value
	07eh -- byte -- COMM3 timeout value
	07fh -- byte -- COMM4 timeout value


port defined:
        serial port

 port 2f8h-2ffh -- serial port 2
 port 3f8h-3ffh -- serial port 1
  X = 3 for port 1
    = 2 for port 2
       i/o address     register selected                 DLAB state
       Xf8h            TX buffer                         0(write)
       Xf8h            RX buffer                         0(read)
       Xf8h            divisor latch LSB                 1
       Xf9h            divisor latch MSB                 1 
       Xf9h            interrupt enable register         0
       Xfah            interrupt identification register
       Xfbh            line control register
       Xfch            modem control register
       Xfdh            line status
       Xfeh            modem status register
       Xffh            reserved

 port Xf8 - transmitter holding register
               bit 0 -- data bit 0
               bit 1 -- data bit 1
               bit 2 -- data bit 2
               bit 3 -- data bit 3
               bit 4 -- data bit 4
               bit 5 -- data bit 5
               bit 6 -- data bit 6
               bit 7 -- data bit 7

 port Xf8 - receiver buffer register
               bit 0 -- data bit 0
               bit 1 -- data bit 1
               bit 2 -- data bit 2
               bit 3 -- data bit 3
               bit 4 -- data bit 4
               bit 5 -- data bit 5
               bit 6 -- data bit 6
               bit 7 -- data bit 7

 port Xf8 - divisor latch LSB
               bit 0 -- bit 0
               bit 1 -- bit 1
               bit 2 -- bit 2
               bit 3 -- bit 3
               bit 4 -- bit 4
               bit 5 -- bit 5
               bit 6 -- bit 6
               bit 7 -- bit 7


 port Xf9 - divisor MSB
               bit 0 -- bit 0
               bit 1 -- bit 1
               bit 2 -- bit 2
               bit 3 -- bit 3
               bit 4 -- bit 4
               bit 5 -- bit 5
               bit 6 -- bit 6
               bit 7 -- bit 7

 port Xf9 - interrupt enable register
               bit 0 -- enable the received-data-available interrupt
               bit 1 -- enable Tx holding register empty interrupt
               bit 2 -- enable receive line status interrupt
               bit 3 -- enable receive modem status interrupt
               bit 4 -- = 0
               bit 5 -- = 0
               bit 6 -- = 0
               bit 7 -- = 0

 port Xfa - interrupt indentification register 
               bit 0 -- = 0 if interrupt pending
               bit 1-2  interrupt ID
               bit 3 -- = 0
               bit 4 -- = 0
               bit 5 -- = 0
               bit 6 -- = 0
               bit 7 -- = 0

bit 1,2  priority  interrupt         interrupt       interrupt
           level   type              source          reset control
    
    1 1  highest   receiver line     overrun error,  reading the line
                   status            parity error,   satus register
                                     framing error,
                                     break interrupt

    1 0  second    received data     receiver data   reading the receiver
                   available         available       buffer register

    0 1  third     transmitter       transmitter     read the IIR (if 
                   holding register  holding rgister source of interrupt)
                   empty             empty           or writing into the
                                                     THR
                                                       
    0 0  fourth    modem status      clear to send,  reading the modem 
                                     data set ready, status register
                                     ring indicator,
                                     received line
                                     signal detect


 port Xfbh - line control register
               bit 0-1         word length (bit)
                         = 0    5
                         = 1    6
                         = 2    7
                         = 3    8
               bit 2 --  number of stop bits
               bit 3 --  parity enable
               bit 4 --  even parity select
               bit 5 --  stuck parity
               bit 6 --  set break
               bit 7 --  divisor latch access bit

 port Xfch - modem control register 
               bit 0 -- data terminal ready
               bit 1 -- request to send
               bit 2 -- out 1
               bit 3 -- out 2
               bit 4 -- loop
               bit 5 -- = 0
               bit 6 -- = 0
               bit 7 -- = 0

 port Xfdh - line status register
               bit 0 -- data ready
               bit 1 -- overrun error
               bit 2 -- parity error
               bit 3 -- framing error
               bit 4 -- break interrupt
               bit 5 -- transmittter holding register empty
               bit 6 -- Tx shift register empty
               bit 7 -- = 0

 port Xfeh - modem status register
               bit 0 -- delta clear to send
               bit 1 -- delta data set ready
               bit 2 -- trailing edge ring indicator
               bit 3 -- delta data carrier detect
               bit 4 -- clear to sent
               bit 5 -- data set ready
               bit 6 -- ring indicator
               bit 7 -- data carrier detect


programming:
	get base port address
	check serial port address (0 means not installed)
	check calling function number (0 - 5)


INITIALIZE THE COMMUNICATIONS PORT
	input:	ah = 0
		al = parameters for initialization

	 7	6	5	4	3	2	 1	 0
	 --- baud rate ---	--parity--    stopbit	-word length-
	 000 - 110		00 - none	0 - 1	10 - 7 bits
	 001 - 150		01 - odd	1 - 2	11 - 8 bits
	 010 - 300		11 - even
	 011 - 600
	 100 - 1200
	 101 - 2400
	 110 - 4800
	 111 - 9600

		dx = device number (0,1,2,3)


	output: ah = status
		ah = line control status as follows:
		   bit 7 = time out
		   bit 6 = trans shift register empty
		   bit 5 = tran holding register empty
		   bit 4 = break detect
		   bit 3 = framing error
		   bit 2 = parity error
		   bit 1 = overrun error
		   bit 0 = data ready
		al = modem status as follows:
		   bit 7 = receved line signal detect
		   bit 6 = rinc indicator
		   bit 5 = data set ready
		   bit 4 = clear to send
		   bit 3 = delta receive line signal detect
		   bit 2 = trailing edge ring detector
		   bit 1 = delta data set ready
		   bit 0 = delta clear to  send


SEND CHARACTER
	input:	ah = 1
		al = character
		dx = device number (0,1,2,3)

	output: al = character
		ah -- bit 7 = 1 -- if character was sent
		 if bit 7 = 0 then bits 6-0 are:
		    bit 6 = trans shift register empty
		    bit 5 = tran holding register empty
		    bit 4 = break detect
		    bit 3 = framing error
		    bit 2 = parity error
		    bit 1 = overrun error
		    bit 0 = data ready

	programming:
		set DTR & RTS on
		test DSR & CTS
		If both on
		   then test THRE
			if not timeout,
			   then send character
			else set timeout flag
		else set timeout flag

Note:
 1. Term definitions:
      DTR  -- Data Terminal Ready
      RTS  -- Request To Send
      DSR  -- Data Set Ready
      CTS  -- Clear To Send
      THRE -- Transmitter Holding Register Empty

 2. The time-out value is approximately 1.2 seconds.


RECEIVE A CHARACTER
	input:	ah = 2
		dx = device number (0,1,2,3)
	output: ah = 0 -- no error
		   al = character received
		ah = error status
		   bit 7 = time out
		   bit 4 = break detect
		   bit 3 = framing error
		   bit 2 = parity error
		   bit 1 = overrun error

	programming:
		set DTR on
		test DSR
		if on
		   then test DR
			if on
			   then receive character
			else set timeout flag
		else set timeout flag

Note:
 1. Term definitions:
      DTR  -- Data Terminal Ready
      DSR  -- Data Set Ready
      DR   -- Data Ready

 2. The time-out value is approximately 1.2 seconds.


RETURN THE COMMUNICATIONS PORT STATUS
	input:	ah = 3
		dx = device number (0,1,2,3)
	output:
		ah = comm line status (Comm status) as follows:
		   bit 7 = time out
		   bit 6 = trans shift register empty
		   bit 5 = tran holding register empty
		   bit 4 = break detect
		   bit 3 = framing error
		   bit 2 = parity error
		   bit 1 = overrun error
		   bit 0 = data ready
		al = modem status as follows:
		   bit 7 = received line signal detect
		   bit 6 = ring indicator
		   bit 5 = data set ready
		   bit 4 = clear to send
		   bit 3 = delta receive line signal detect
		   bit 2 = trailing edge ring detector
		   bit 1 = delta data set ready
		   bit 0 = delta clear to  send

		dx = RS232 card selector (0,1)

	programming:
		read line status & modem status from related i/o port


EXTENDED INITIALIZE
	input:
		ah = 4
		al - break
			00 - no break
			01 - break
		bh - parity
			00 - none
			01 - odd
			02 - even
			03 - stick parity odd
			04 - stick parity even
		bl - stop bit
			00 - one
			01 - two if 6-, 7-, or 8-bit word length
			   - one-and-one-half if 5-bit word length
		ch - word length
			= 00h - 5 bits
			= 01h - 6 bits
			= 02h - 7 bits
			= 03h - 8 bits
		cl - baud rate
			= 00h - 110 baud
			= 01h - 150 baud
			= 02h - 300 baud
			= 03h - 600 baud
			= 04h - 1200 baud
			= 05h - 2400 baud
			= 06h - 4800 baud
			= 07h - 9600 baud
			= 08h - 19200 baud

	output:
		al - modem status
		ah - line status


EXTENDED COMMUNICATIONS PORT CONTROL
	input:
	  ah = 5
		al = 0 - read modem control register
		dx = device number (0,1,2,3)
		output:
			bl - modem control register
				bit 0 = 1 - data terminal ready
				bit 1 = 1 - request to send
				bit 2 = 1 - out1
				bit 3 = 1 - out2
				bit 4 = 1 - loop
				bit 5 - 7 - reserved

		al = 1 - write modem control register
		dx = device number (0,1,2,3)
		bl = data to write to modem control register
		   bit 0 = 1 - data terminal ready
		   bit 1 = 1 - request to send
		   bit 2 = 1 - out1
		   bit 3 = 1 - out2
		   bit 4 = 1 - loop
		   bit 5 - 7 - reserved

	output:
		al - modem status
		ah - line control status

	programming:
		check al
		if al = 0 then read from modem control port into bl
		if al = 1 then write bl to port
		else null return


RESERVED
	input:
	        ah = 006h - 0ffh 

	output:
                none (ax is changed)


