
                 Specification of ACER Setup Routine
                 ------------------------------------

Description:
        This routine is a utility in ROM BIOS. It sets up the system 
        configuration. This routine is called by INT 9h when user press 
        Ctrl-Alt-Esc key at the same time.
        The main menu of the setup service routine includes three items: 
        Power on password, Hard disk physical format and Set up system 
        configuration.

   A.
        The first item sets up the user password when the user has no 
        password. Once the password has been set up, this function must 
        tell the user that he can't update password through this 
        item and he can set the new or clear old password during
        power-on time only.
        
        NOTE: Reference CMOS data 3ch-3fh

   B.
        The second item performs hard disk physical formatting. 
        The user must input bad cylinder, head, and interleave 
        number before doing physical formatting.

   C.
        The third item sets up the system configuration and 
        the second item physically formats the hard disk.
        
        System configuration setup items are seperated into two screen:
        one for common and the other for special items for each system.

        The first screen of the system configuration includes the following
        items:

               1. Date
               2. Time
               3. Diskette type and No.
               4. Hard disk type and No.
               5. Base and extended memory size
               6. Primary display
               7. Coprocesser status
               8. CPU speed/cache


Function:
   1. DATE :
         This item sets the system date which includes the day, month and 
         year. The format of date on screen is mm/dd/yy.
         NOTE: Reference CMOS data 07h - 09h

   2. TIME :
         This item sets the system time which includes hour, minute and 
         second. The format of date on screen is hr/min/sec.
         NOTE: Reference CMOS data 00h,02h,04h

   3. DISKETTE TYPE AND No.
         This item supports the set diskette number (driver A and B) and
         diskette type (360K, 720K, 1.2Mb and 1.44Mb).
         NOTE: Reference CMOS data 10h

   4. HARD DISK TYPE AND No.
         This item supports the set hard disk number (driver C and D) and
         hard disk type. The user can select type 255 to define his own
         hard disk cylinder, sector, head, ship landing zone and pre_com.
         NOTE: Reference CMOS data 12h,19h,1ah 
               user define type 1eh - 2dh

   5. BASE AND EXTENDED MEMORY SIZE
         These two items comprise the system memory map: base memory
         size and extended memory above 1M. This item displays the
         memory size only.
         NOTE: Reference CMOS data 15h - 18h

   6. PRIMARY DISPLAY
         This item shows which primary display type is attached and checked 
         during POST routine. Do not allow the user to select this display 
         type.
         NOTE: Reference CMOS data 14h bit 5,4

   7. COPROCESSER STATUS
         This item shows the presence or absence of a coprocesser.
         NOTE: Reference CMOS data 14h bit 1

   8. CPU SPEED/CACHE
         This item selects the CPU speed and cache.          
         The speed value depends on each system. This function has three
         options: 1. high and low speed only (for 910, 915) 
                  2. multiscale speed (for 1100/20/16)
                  3. multiscale speed and cache (for 1100/25/20c/33)
         NOTE: Reference CMOS data 39h bit 7-4
         

        The second screen of the system configuration contains some special
        items for each system. Each item is described below:

               1. Shadow RAM (system and video)
               2. Printer port Status
               3. COM port Status
               4. WDD/FDD write protection
               5. COM1/Console Status
               6. EMS Status
               7. Power Saving 
               8. MODEM Status


Function:

   1. SHADOW RAM BIOS
         This item selects the BIOS as shadow ram. The BIOS includes
         system BIOS and video BIOS. 
         NOTE: Reference CMOS data 39h bit 1,0

   2. PRINTER PORT  
         This item enables/disables the printer port and also assigns
         printer port. The assigned printer port can be 3BCh, 378h, or
         278h.
         NOTE: Reference CMOS data 37h bit 1,0

   3. COM PORT
         If an add-on COM port exists, then POST routine disables the 
         built-in COM port. This item displays built-in COM port status. 
         NOTE: Reference CMOS data 37h bit 3,2

   4. WDD AND FDD WRITE PROTECT
         User uses this item to select whether WDD and FDD are write 
         protected or not to prevent VIRUS intrusion. 
         NOTE: Reference CMOS data 39h bit 3,2

   5. COM1/CONSOLE
         If the system does not have any display, user can select COM1 port
         as the display, COM1 port information will be set before the
         system operates (i.e, baud rate, parity, data bit and stop bits).
         NOTE: Reference CMOS data 34h     



   6. EMS STATUS
         This item enables/disables the EMS, and for NEAT system lets the 
         user set EMS size.
         NOTE: Reference CMOS data 37h bit 7-4

   7. POWER SAVING
         This item is supported for the laptop system only. There are LCD 
         display power saving, hard disk power saving and system power 
         saving. Set the LCD power auto inactive time in the range of 1-63
         min. Set hard disk power auto inactive time in the range of 1-18.
         Set system power auto inactive time in the range of 1-23 min.
         NOTE: Reference CMOS data 1bh-1dh
         
   8. MODEM STATUS
         This item is supported for the laptop system only. The user can 
         select the MODEM power on or off.
         NOTE: Reference CMOS data 35h bit 0


Key definition:
        up and down arrow key    -- to select main item
        right and left arrow key -- to select subitem
        pgup and pgdn key        -- to select another page
        ESC key                  -- exit setup routine
        F1 key                   -- show help message      


   NOTE:
      1. User should use right and left arrow key to select optional 
         subitem. Subitem should not be keyed in.

      2. When user presses the 'ESC' key to confirm his modification, 
         CMOS is not updated until the user confirms his modification.

